@@ -1,4 +1,3 @@
-/* $Id: rcfile.c 5679 2016-02-25 21:04:45Z astyanax $ */
 /**************************************************************************
  *   rcfile.c                                                             *
  *                                                                        *
@@ -120,9 +119,12 @@ static size_t lineno = 0;
 static char *nanorc = NULL;
 	/* The path to the rcfile we're parsing. */
 #ifndef DISABLE_COLOR
-static syntaxtype *endsyntax = NULL;
-	/* The end of the list of syntaxes. */
-static colortype *endcolor = NULL;
+static bool opensyntax = FALSE;
+	/* Whether we're allowed to add to the last syntax.  When a file ends,
+	 * or when a new syntax command is seen, this bool becomes FALSE. */
+static syntaxtype *live_syntax;
+	/* The syntax that is currently being parsed. */
+static colortype *lastcolor = NULL;
 	/* The end of the color list for the current syntax. */
 #endif
 
@@ -210,15 +212,16 @@ char *parse_argument(char *ptr)
 }
 
 #ifndef DISABLE_COLOR
-/* Parse the next regex string from the line at ptr, and return it. */
+/* Pass over the current regex string in the line starting at ptr,
+ * null-terminate it, and return a pointer to the /next/ word. */
 char *parse_next_regex(char *ptr)
 {
     assert(ptr != NULL);
 
-    /* Continue until the end of the line, or a " followed by a space, a
-     * blank character, or \0. */
-    while ((*ptr != '"' || (!isblank(*(ptr + 1)) &&
-	*(ptr + 1) != '\0')) && *ptr != '\0')
+    /* Continue until the end of line, or until a " followed by a
+     * blank character or the end of line. */
+    while (*ptr != '\0' && (*ptr != '"' ||
+		(*(ptr + 1) != '\0' && !isblank(*(ptr + 1)))))
 	ptr++;
 
     assert(*ptr == '"' || *ptr == '\0');
@@ -239,12 +242,12 @@ char *parse_next_regex(char *ptr)
 }
 
 /* Compile the regular expression regex to see if it's valid.  Return
- * TRUE if it is, or FALSE otherwise. */
-bool nregcomp(const char *regex, int eflags)
+ * TRUE if it is, and FALSE otherwise. */
+bool nregcomp(const char *regex, int compile_flags)
 {
     regex_t preg;
     const char *r = fixbounds(regex);
-    int rc = regcomp(&preg, r, REG_EXTENDED | eflags);
+    int rc = regcomp(&preg, r, compile_flags);
 
     if (rc != 0) {
 	size_t len = regerror(rc, &preg, NULL, 0);
@@ -259,133 +262,75 @@ bool nregcomp(const char *regex, int eflags)
     return (rc == 0);
 }
 
-/* Parse the next syntax string from the line at ptr, and add it to the
- * global list of color syntaxes. */
+/* Parse the next syntax name and its possible extension regexes from the
+ * line at ptr, and add it to the global linked list of color syntaxes. */
 void parse_syntax(char *ptr)
 {
-    const char *fileregptr = NULL, *nameptr = NULL;
-    syntaxtype *tmpsyntax, *prev_syntax;
-    regexlisttype *endext = NULL;
-	/* The end of the extensions list for this syntax. */
+    char *nameptr;
+	/* A pointer to what should be the name of the syntax. */
+
+    opensyntax = FALSE;
 
     assert(ptr != NULL);
 
-    if (*ptr == '\0') {
+    /* Check that the syntax name is not empty. */
+    if (*ptr == '\0' || (*ptr == '"' &&
+			(*(ptr + 1) == '\0' || *(ptr + 1) == '"'))) {
 	rcfile_error(N_("Missing syntax name"));
 	return;
     }
 
-    if (*ptr != '"') {
-	rcfile_error(
-		N_("Regex strings must begin and end with a \" character"));
+    nameptr = ++ptr;
+    ptr = parse_next_word(ptr);
+
+    /* Check that the name starts and ends with a double quote. */
+    if (*(nameptr - 1) != '\x22' || nameptr[strlen(nameptr) - 1] != '\x22') {
+	rcfile_error(N_("A syntax name must be quoted"));
 	return;
     }
 
-    ptr++;
+    /* Strip the end quote. */
+    nameptr[strlen(nameptr) - 1] = '\0';
 
-    nameptr = ptr;
-    ptr = parse_next_regex(ptr);
-
-    if (ptr == NULL)
+    /* Redefining the "none" syntax is not allowed. */
+    if (strcmp(nameptr, "none") == 0) {
+	rcfile_error(N_("The \"none\" syntax is reserved"));
 	return;
-
-    /* Search for a duplicate syntax name.  If we find one, free it, so
-     * that we always use the last syntax with a given name. */
-    prev_syntax = NULL;
-    for (tmpsyntax = syntaxes; tmpsyntax != NULL;
-	tmpsyntax = tmpsyntax->next) {
-	if (strcmp(nameptr, tmpsyntax->desc) == 0) {
-	    syntaxtype *old_syntax = tmpsyntax;
-
-	    if (endsyntax == tmpsyntax)
-		endsyntax = prev_syntax;
-
-	    tmpsyntax = tmpsyntax->next;
-	    if (prev_syntax != NULL)
-		prev_syntax->next = tmpsyntax;
-	    else
-		syntaxes = tmpsyntax;
-
-	    free(old_syntax->desc);
-	    free(old_syntax);
-	    break;
-	}
-	prev_syntax = tmpsyntax;
     }
 
-    if (syntaxes == NULL) {
-	syntaxes = (syntaxtype *)nmalloc(sizeof(syntaxtype));
-	endsyntax = syntaxes;
-    } else {
-	endsyntax->next = (syntaxtype *)nmalloc(sizeof(syntaxtype));
-	endsyntax = endsyntax->next;
-#ifdef DEBUG
-	fprintf(stderr, "Adding new syntax after first one\n");
-#endif
-    }
+    /* Initialize a new syntax struct. */
+    live_syntax = (syntaxtype *)nmalloc(sizeof(syntaxtype));
+    live_syntax->name = mallocstrcpy(NULL, nameptr);
+    live_syntax->extensions = NULL;
+    live_syntax->headers = NULL;
+    live_syntax->magics = NULL;
+    live_syntax->linter = NULL;
+    live_syntax->formatter = NULL;
+    live_syntax->comment = NULL;
+    live_syntax->color = NULL;
+    lastcolor = NULL;
+    live_syntax->nmultis = 0;
 
-    endsyntax->desc = mallocstrcpy(NULL, nameptr);
-    endsyntax->color = NULL;
-    endcolor = NULL;
-    endsyntax->extensions = NULL;
-    endsyntax->headers = NULL;
-    endsyntax->magics = NULL;
-    endsyntax->next = NULL;
-    endsyntax->nmultis = 0;
-    endsyntax->linter = NULL;
-    endsyntax->formatter = NULL;
+    /* Hook the new syntax in at the top of the list. */
+    live_syntax->next = syntaxes;
+    syntaxes = live_syntax;
+
+    opensyntax = TRUE;
 
 #ifdef DEBUG
     fprintf(stderr, "Starting a new syntax type: \"%s\"\n", nameptr);
 #endif
 
-    /* The "none" syntax is the same as not having a syntax at all, so
-     * we can't assign any extensions or colors to it. */
-    if (strcmp(endsyntax->desc, "none") == 0) {
-	rcfile_error(N_("The \"none\" syntax is reserved"));
-	return;
-    }
-
     /* The default syntax should have no associated extensions. */
-    if (strcmp(endsyntax->desc, "default") == 0 && *ptr != '\0') {
+    if (strcmp(live_syntax->name, "default") == 0 && *ptr != '\0') {
 	rcfile_error(
-		N_("The \"default\" syntax must take no extensions"));
+		N_("The \"default\" syntax does not accept extensions"));
 	return;
     }
 
-    /* Now load the extension regexes into their part of the struct. */
-    while (*ptr != '\0') {
-	regexlisttype *newext;
-
-	while (*ptr != '"' && *ptr != '\0')
-	    ptr++;
-
-	if (*ptr == '\0')
-	    return;
-
-	ptr++;
-
-	fileregptr = ptr;
-	ptr = parse_next_regex(ptr);
-	if (ptr == NULL)
-	    break;
-
-	newext = (regexlisttype *)nmalloc(sizeof(regexlisttype));
-
-	/* Save the extension regex if it's valid. */
-	if (nregcomp(fileregptr, REG_NOSUB)) {
-	    newext->ext_regex = mallocstrcpy(NULL, fileregptr);
-	    newext->ext = NULL;
-
-	    if (endext == NULL)
-		endsyntax->extensions = newext;
-	    else
-		endext->next = newext;
-	    endext = newext;
-	    endext->next = NULL;
-	} else
-	    free(newext);
-    }
+    /* If there seem to be extension regexes, pick them up. */
+    if (*ptr != '\0')
+	grab_and_store("extension", ptr, &live_syntax->extensions);
 }
 #endif /* !DISABLE_COLOR */
 
@@ -569,10 +514,9 @@ void parse_binding(char *ptr, bool dobind)
     free(keycopy);
 }
 
-
 #ifndef DISABLE_COLOR
-/* Read and parse additional syntax files. */
-static void _parse_include(char *file)
+/* Read and parse one included syntax file. */
+static void parse_one_include(char *file)
 {
     struct stat rcinfo;
     FILE *rcstream;
@@ -610,7 +554,8 @@ static void _parse_include(char *file)
     parse_rcfile(rcstream, TRUE);
 }
 
-void parse_include(char *ptr)
+/* Expand globs in the passed name, and parse the resultant files. */
+void parse_includes(char *ptr)
 {
     char *option, *nanorc_save = nanorc, *expanded;
     size_t lineno_save = lineno, i;
@@ -626,16 +571,15 @@ void parse_include(char *ptr)
 
     if (glob(expanded, GLOB_ERR|GLOB_NOSORT, NULL, &files) == 0) {
 	for (i = 0; i < files.gl_pathc; ++i)
-	    _parse_include(files.gl_pathv[i]);
-    } else {
+	    parse_one_include(files.gl_pathv[i]);
+    } else
 	rcfile_error(_("Error expanding %s: %s"), option,
 		strerror(errno));
-    }
 
     globfree(&files);
     free(expanded);
 
-    /* We're done with the new syntax file.  Restore the original
+    /* We're done with the included file(s).  Restore the original
      * filename and line number position. */
     nanorc = nanorc_save;
     lineno = lineno_save;
@@ -645,8 +589,6 @@ void parse_include(char *ptr)
  * and set bright to TRUE if that color is bright. */
 short color_to_short(const char *colorname, bool *bright)
 {
-    short mcolor = -1;
-
     assert(colorname != NULL && bright != NULL);
 
     if (strncasecmp(colorname, "bright", 6) == 0) {
@@ -655,35 +597,34 @@ short color_to_short(const char *colorname, bool *bright)
     }
 
     if (strcasecmp(colorname, "green") == 0)
-	mcolor = COLOR_GREEN;
+	return COLOR_GREEN;
     else if (strcasecmp(colorname, "red") == 0)
-	mcolor = COLOR_RED;
+	return COLOR_RED;
     else if (strcasecmp(colorname, "blue") == 0)
-	mcolor = COLOR_BLUE;
+	return COLOR_BLUE;
     else if (strcasecmp(colorname, "white") == 0)
-	mcolor = COLOR_WHITE;
+	return COLOR_WHITE;
     else if (strcasecmp(colorname, "yellow") == 0)
-	mcolor = COLOR_YELLOW;
+	return COLOR_YELLOW;
     else if (strcasecmp(colorname, "cyan") == 0)
-	mcolor = COLOR_CYAN;
+	return COLOR_CYAN;
     else if (strcasecmp(colorname, "magenta") == 0)
-	mcolor = COLOR_MAGENTA;
+	return COLOR_MAGENTA;
     else if (strcasecmp(colorname, "black") == 0)
-	mcolor = COLOR_BLACK;
-    else
-	rcfile_error(N_("Color \"%s\" not understood.\n"
+	return COLOR_BLACK;
+
+    rcfile_error(N_("Color \"%s\" not understood.\n"
 		"Valid colors are \"green\", \"red\", \"blue\",\n"
 		"\"white\", \"yellow\", \"cyan\", \"magenta\" and\n"
 		"\"black\", with the optional prefix \"bright\"\n"
 		"for foreground colors."), colorname);
-
-    return mcolor;
+    return -1;
 }
 
 /* Parse the color string in the line at ptr, and add it to the current
- * file's associated colors.  If icase is TRUE, treat the color string
- * as case insensitive. */
-void parse_colors(char *ptr, bool icase)
+ * file's associated colors.  rex_flags are the regex compilation flags
+ * to use, excluding or including REG_ICASE for case (in)sensitivity. */
+void parse_colors(char *ptr, int rex_flags)
 {
     short fg, bg;
     bool bright = FALSE;
@@ -691,9 +632,10 @@ void parse_colors(char *ptr, bool icase)
 
     assert(ptr != NULL);
 
-    if (syntaxes == NULL) {
+    if (!opensyntax) {
 	rcfile_error(
-		N_("Cannot add a color command without a syntax command"));
+		N_("A '%s' command requires a preceding 'syntax' command"),
+		"color");
 	return;
     }
 
@@ -708,7 +650,7 @@ void parse_colors(char *ptr, bool icase)
 	return;
 
     if (*ptr == '\0') {
-	rcfile_error(N_("Missing regex string"));
+	rcfile_error(N_("Missing regex string after '%s' command"), "color");
 	return;
     }
 
@@ -717,10 +659,10 @@ void parse_colors(char *ptr, bool icase)
     while (ptr != NULL && *ptr != '\0') {
 	colortype *newcolor;
 	    /* The container for a color plus its regexes. */
-	bool cancelled = FALSE;
-	    /* The start expression was bad. */
+	bool goodstart;
+	    /* Whether the start expression was valid. */
 	bool expectend = FALSE;
-	    /* Do we expect an end= line? */
+	    /* Whether to expect an end= line. */
 
 	if (strncasecmp(ptr, "start=", 6) == 0) {
 	    ptr += 6;
@@ -734,22 +676,22 @@ void parse_colors(char *ptr, bool icase)
 	    continue;
 	}
 
-	ptr++;
-
-	fgstr = ptr;
+	fgstr = ++ptr;
 	ptr = parse_next_regex(ptr);
 	if (ptr == NULL)
 	    break;
 
-	newcolor = (colortype *)nmalloc(sizeof(colortype));
+	goodstart = nregcomp(fgstr, rex_flags);
+
+	/* If the starting regex is valid, initialize a new color struct,
+	 * and hook it in at the tail of the linked list. */
+	if (goodstart) {
+	    newcolor = (colortype *)nmalloc(sizeof(colortype));
 
-	/* Save the starting regex string if it's valid, and set up the
-	 * color information. */
-	if (nregcomp(fgstr, icase ? REG_ICASE : 0)) {
 	    newcolor->fg = fg;
 	    newcolor->bg = bg;
 	    newcolor->bright = bright;
-	    newcolor->icase = icase;
+	    newcolor->rex_flags = rex_flags;
 
 	    newcolor->start_regex = mallocstrcpy(NULL, fgstr);
 	    newcolor->start = NULL;
@@ -759,61 +701,48 @@ void parse_colors(char *ptr, bool icase)
 
 	    newcolor->next = NULL;
 
-	    if (endcolor == NULL) {
-		endsyntax->color = newcolor;
 #ifdef DEBUG
-		fprintf(stderr, "Starting a new colorstring for fg %hd, bg %hd\n", fg, bg);
+	    fprintf(stderr, "Adding an entry for fg %hd, bg %hd\n", fg, bg);
 #endif
-	    } else {
-#ifdef DEBUG
-		fprintf(stderr, "Adding new entry for fg %hd, bg %hd\n", fg, bg);
-#endif
-		/* Need to recompute endcolor now so we can extend
-		 * colors to syntaxes. */
-		for (endcolor = endsyntax->color; endcolor->next != NULL; endcolor = endcolor->next)
-		    ;
-		endcolor->next = newcolor;
-	    }
+	    if (lastcolor == NULL)
+		live_syntax->color = newcolor;
+	    else
+		lastcolor->next = newcolor;
 
-	    endcolor = newcolor;
-	} else {
-	    free(newcolor);
-	    cancelled = TRUE;
+	    lastcolor = newcolor;
 	}
 
-	if (expectend) {
-	    if (ptr == NULL || strncasecmp(ptr, "end=", 4) != 0) {
-		rcfile_error(
-			N_("\"start=\" requires a corresponding \"end=\""));
-		return;
-	    }
-	    ptr += 4;
-	    if (*ptr != '"') {
-		rcfile_error(
-			N_("Regex strings must begin and end with a \" character"));
-		continue;
-	    }
+	if (!expectend)
+	    continue;
 
-	    ptr++;
-
-	    fgstr = ptr;
-	    ptr = parse_next_regex(ptr);
-	    if (ptr == NULL)
-		break;
-
-	    /* If the start regex was invalid, skip past the end regex
-	     * to stay in sync. */
-	    if (cancelled)
-		continue;
-
-	    /* Save the ending regex string if it's valid. */
-	    newcolor->end_regex = (nregcomp(fgstr, icase ? REG_ICASE :
-		0)) ? mallocstrcpy(NULL, fgstr) : NULL;
-
-	    /* Lame way to skip another static counter. */
-	    newcolor->id = endsyntax->nmultis;
-	    endsyntax->nmultis++;
+	if (ptr == NULL || strncasecmp(ptr, "end=", 4) != 0) {
+	    rcfile_error(N_("\"start=\" requires a corresponding \"end=\""));
+	    return;
 	}
+
+	ptr += 4;
+	if (*ptr != '"') {
+	    rcfile_error(N_("Regex strings must begin and end with a \" character"));
+	    continue;
+	}
+
+	fgstr = ++ptr;
+	ptr = parse_next_regex(ptr);
+	if (ptr == NULL)
+	    break;
+
+	/* If the start regex was invalid, skip past the end regex
+	 * to stay in sync. */
+	if (!goodstart)
+	    continue;
+
+	/* If it's valid, save the ending regex string. */
+	if (nregcomp(fgstr, rex_flags))
+	    newcolor->end_regex = mallocstrcpy(NULL, fgstr);
+
+	/* Lame way to skip another static counter. */
+	newcolor->id = live_syntax->nmultis;
+	live_syntax->nmultis++;
     }
 }
 
@@ -855,177 +784,112 @@ bool parse_color_names(char *combostr, short *fg, short *bg, bool *bright)
     return TRUE;
 }
 
-/* Parse the header-line regexes that may influence the choice of syntax. */
-void parse_header_exp(char *ptr)
+/* Read regex strings enclosed in double quotes from the line pointed at
+ * by ptr, and store them quoteless in the passed storage place. */
+void grab_and_store(const char *kind, char *ptr, regexlisttype **storage)
 {
-    regexlisttype *endheader = NULL;
+    regexlisttype *lastthing;
 
-    assert(ptr != NULL);
-
-    if (syntaxes == NULL) {
+    if (!opensyntax) {
 	rcfile_error(
-		N_("Cannot add a header regex without a syntax command"));
+		N_("A '%s' command requires a preceding 'syntax' command"), kind);
+	return;
+    }
+
+    /* The default syntax doesn't take any file matching stuff. */
+    if (strcmp(live_syntax->name, "default") == 0 && *ptr != '\0') {
+	rcfile_error(
+		N_("The \"default\" syntax does not accept '%s' regexes"), kind);
 	return;
     }
 
     if (*ptr == '\0') {
-	rcfile_error(N_("Missing regex string"));
+	rcfile_error(N_("Missing regex string after '%s' command"), kind);
 	return;
     }
 
+    lastthing = *storage;
+
+    /* If there was an earlier command, go to the last of those regexes. */
+    while (lastthing != NULL && lastthing->next != NULL)
+	lastthing = lastthing->next;
+
+    /* Now gather any valid regexes and add them to the linked list. */
     while (*ptr != '\0') {
 	const char *regexstring;
-	regexlisttype *newheader;
+	regexlisttype *newthing;
 
 	if (*ptr != '"') {
 	    rcfile_error(
 		N_("Regex strings must begin and end with a \" character"));
-	    ptr = parse_next_regex(ptr);
-	    continue;
+	    return;
 	}
 
-	ptr++;
-
-	regexstring = ptr;
+	regexstring = ++ptr;
 	ptr = parse_next_regex(ptr);
 	if (ptr == NULL)
-	    break;
-
-	newheader = (regexlisttype *)nmalloc(sizeof(regexlisttype));
-
-	/* Save the regex string if it's valid. */
-	if (nregcomp(regexstring, 0)) {
-	    newheader->ext_regex = mallocstrcpy(NULL, regexstring);
-	    newheader->ext = NULL;
-
-	    if (endheader == NULL)
-		endsyntax->headers = newheader;
-	    else
-		endheader->next = newheader;
-	    endheader = newheader;
-	    endheader->next = NULL;
-	} else
-	    free(newheader);
-    }
-}
-
-#ifdef HAVE_LIBMAGIC
-/* Parse the magic regexes that may influence the choice of syntax. */
-void parse_magic_exp(char *ptr)
-{
-    regexlisttype *endmagic = NULL;
-
-    assert(ptr != NULL);
-
-    if (syntaxes == NULL) {
-	rcfile_error(
-		N_("Cannot add a magic string regex without a syntax command"));
-	return;
-    }
-
-    if (*ptr == '\0') {
-	rcfile_error(N_("Missing magic string name"));
-	return;
-    }
-
-    if (*ptr != '"') {
-	rcfile_error(
-		N_("Regex strings must begin and end with a \" character"));
-	return;
-    }
-
-#ifdef DEBUG
-    fprintf(stderr, "Starting a magic type: \"%s\"\n", ptr);
-#endif
-
-    /* Now load the magic regexes into their part of the struct. */
-    while (*ptr != '\0') {
-	const char *regexstring;
-	regexlisttype *newmagic;
-
-	while (*ptr != '"' && *ptr != '\0')
-	    ptr++;
-
-	if (*ptr == '\0')
 	    return;
 
-	ptr++;
+	/* If the regex string is malformed, skip it. */
+	if (!nregcomp(regexstring, NANO_REG_EXTENDED | REG_NOSUB))
+	    continue;
 
-	regexstring = ptr;
-	ptr = parse_next_regex(ptr);
-	if (ptr == NULL)
-	    break;
+	/* Copy the regex into a struct, and hook this in at the end. */
+	newthing = (regexlisttype *)nmalloc(sizeof(regexlisttype));
+	newthing->full_regex = mallocstrcpy(NULL, regexstring);
+	newthing->next = NULL;
 
-	newmagic = (regexlisttype *)nmalloc(sizeof(regexlisttype));
+	if (lastthing == NULL)
+	    *storage = newthing;
+	else
+	    lastthing->next = newthing;
 
-	/* Save the regex string if it's valid. */
-	if (nregcomp(regexstring, REG_NOSUB)) {
-	    newmagic->ext_regex = mallocstrcpy(NULL, regexstring);
-	    newmagic->ext = NULL;
-
-	    if (endmagic == NULL)
-		endsyntax->magics = newmagic;
-	    else
-		endmagic->next = newmagic;
-	    endmagic = newmagic;
-	    endmagic->next = NULL;
-	} else
-	    free(newmagic);
+	lastthing = newthing;
     }
 }
-#endif /* HAVE_LIBMAGIC */
 
-/* Parse the linter requested for this syntax. */
-void parse_linter(char *ptr)
+/* Parse and store the name given after a linter/formatter command. */
+void pick_up_name(const char *kind, char *ptr, char **storage)
 {
     assert(ptr != NULL);
 
-    if (syntaxes == NULL) {
+    if (!opensyntax) {
 	rcfile_error(
-		N_("Cannot add a linter without a syntax command"));
+		N_("A '%s' command requires a preceding 'syntax' command"), kind);
 	return;
     }
 
     if (*ptr == '\0') {
-	rcfile_error(N_("Missing linter command"));
+	rcfile_error(N_("Missing command after '%s'"), kind);
 	return;
     }
 
-    free(endsyntax->linter);
+    free(*storage);
 
-    /* Let them unset the linter by using "". */
+    /* Allow unsetting the command by using an empty string. */
     if (!strcmp(ptr, "\"\""))
-	endsyntax->linter = NULL;
-    else
-	endsyntax->linter = mallocstrcpy(NULL, ptr);
-}
-
-#ifndef DISABLE_SPELLER
-/* Parse the formatter requested for this syntax. */
-void parse_formatter(char *ptr)
-{
-    assert(ptr != NULL);
-
-    if (syntaxes == NULL) {
-	rcfile_error(
-		N_("Cannot add formatter without a syntax command"));
-	return;
+	*storage = NULL;
+    else if (*ptr == '"') {
+	*storage = mallocstrcpy(NULL, ++ptr);
+	char* q = *storage;
+	char* p = *storage;
+	/* Snip out the backslashes of escaped characters. */
+	while (*p != '"') {
+	    if (*p == '\0') {
+		rcfile_error(N_("Argument of '%s' lacks closing \""), kind);
+		free(*storage);
+		*storage = NULL;
+		return;
+	    } else if (*p == '\\' && *(p + 1) != '\0') {
+		p++;
+	    }
+	    *q++ = *p++;
+	}
+	*q = '\0';
     }
-
-    if (*ptr == '\0') {
-	rcfile_error(N_("Missing formatter command"));
-	return;
-    }
-
-    free(endsyntax->formatter);
-
-    /* Let them unset the formatter by using "". */
-    if (!strcmp(ptr, "\"\""))
-	endsyntax->formatter = NULL;
     else
-	endsyntax->formatter = mallocstrcpy(NULL, ptr);
+	*storage = mallocstrcpy(NULL, ptr);
 }
-#endif /* !DISABLE_SPELLER */
 #endif /* !DISABLE_COLOR */
 
 /* Check whether the user has unmapped every shortcut for a
@@ -1044,10 +908,10 @@ static void check_vitals_mapped(void)
 		const sc *s = first_sc_for(inmenus[v], f->scfunc);
 		if (!s) {
 		    fprintf(stderr, _("Fatal error: no keys mapped for function "
-				     "\"%s\".  Exiting.\n"), f->desc);
+					"\"%s\".  Exiting.\n"), f->desc);
 		    fprintf(stderr, _("If needed, use nano with the -I option "
-				     "to adjust your nanorc settings.\n"));
-		     exit(1);
+					"to adjust your nanorc settings.\n"));
+		    exit(1);
 		}
 		break;
 	    }
@@ -1057,7 +921,7 @@ static void check_vitals_mapped(void)
 
 /* Parse the rcfile, once it has been opened successfully at rcstream,
  * and close it afterwards.  If syntax_only is TRUE, only allow the file
- * to contain color syntax commands: syntax, color, and icolor. */
+ * to contain color syntax commands. */
 void parse_rcfile(FILE *rcstream
 #ifndef DISABLE_COLOR
 	, bool syntax_only
@@ -1067,9 +931,6 @@ void parse_rcfile(FILE *rcstream
     char *buf = NULL;
     ssize_t len;
     size_t n = 0;
-#ifndef DISABLE_COLOR
-    syntaxtype *end_syn_save = NULL;
-#endif
 
     while ((len = getline(&buf, &n, rcstream)) > 0) {
 	char *ptr, *keyword, *option;
@@ -1097,81 +958,79 @@ void parse_rcfile(FILE *rcstream
 #ifndef DISABLE_COLOR
 	/* Handle extending first... */
 	if (strcasecmp(keyword, "extendsyntax") == 0) {
+	    syntaxtype *sint;
 	    char *syntaxname = ptr;
-	    syntaxtype *ts = NULL;
 
 	    ptr = parse_next_word(ptr);
-	    for (ts = syntaxes; ts != NULL; ts = ts->next)
-		if (!strcmp(ts->desc, syntaxname))
+
+	    for (sint = syntaxes; sint != NULL; sint = sint->next)
+		if (!strcmp(sint->name, syntaxname))
 		    break;
 
-	    if (ts == NULL) {
-		rcfile_error(N_("Could not find syntax \"%s\" to extend"), syntaxname);
+	    if (sint == NULL) {
+		rcfile_error(N_("Could not find syntax \"%s\" to extend"),
+				syntaxname);
+		opensyntax = FALSE;
 		continue;
-	    } else {
-		end_syn_save = endsyntax;
-		endsyntax = ts;
-		keyword = ptr;
-		ptr = parse_next_word(ptr);
 	    }
+
+	    live_syntax = sint;
+	    opensyntax = TRUE;
+
+	    /* Refind the tail of the color list for this syntax. */
+	    lastcolor = sint->color;
+	    if (lastcolor != NULL)
+		while (lastcolor->next != NULL)
+		    lastcolor = lastcolor->next;
+
+	    keyword = ptr;
+	    ptr = parse_next_word(ptr);
 	}
-#endif
 
 	/* Try to parse the keyword. */
-	if (strcasecmp(keyword, "set") == 0) {
-#ifndef DISABLE_COLOR
-	    if (syntax_only)
-		rcfile_error(
-			N_("Command \"%s\" not allowed in included file"),
-			keyword);
-	    else
-#endif
-		set = 1;
-	} else if (strcasecmp(keyword, "unset") == 0) {
-#ifndef DISABLE_COLOR
-	    if (syntax_only)
-		rcfile_error(
-			N_("Command \"%s\" not allowed in included file"),
-			keyword);
-	    else
-#endif
-		set = -1;
-	}
-#ifndef DISABLE_COLOR
-	else if (strcasecmp(keyword, "include") == 0) {
-	    if (syntax_only)
-		rcfile_error(
-			N_("Command \"%s\" not allowed in included file"),
-			keyword);
-	    else
-		parse_include(ptr);
-	} else if (strcasecmp(keyword, "syntax") == 0) {
-	    if (endsyntax != NULL && endcolor == NULL)
+	if (strcasecmp(keyword, "syntax") == 0) {
+	    if (opensyntax && lastcolor == NULL)
 		rcfile_error(N_("Syntax \"%s\" has no color commands"),
-			endsyntax->desc);
+				live_syntax->name);
 	    parse_syntax(ptr);
 	}
+	else if (strcasecmp(keyword, "header") == 0)
+	    grab_and_store("header", ptr, &live_syntax->headers);
 	else if (strcasecmp(keyword, "magic") == 0)
 #ifdef HAVE_LIBMAGIC
-	    parse_magic_exp(ptr);
+	    grab_and_store("magic", ptr, &live_syntax->magics);
+#else
+	    ;
+#endif
+	else if (strcasecmp(keyword, "comment") == 0)
+#ifdef ENABLE_COMMENT
+	    pick_up_name("comment", ptr, &live_syntax->comment);
 #else
 	    ;
 #endif
-	else if (strcasecmp(keyword, "header") == 0)
-	    parse_header_exp(ptr);
 	else if (strcasecmp(keyword, "color") == 0)
-	    parse_colors(ptr, FALSE);
+	    parse_colors(ptr, NANO_REG_EXTENDED);
 	else if (strcasecmp(keyword, "icolor") == 0)
-	    parse_colors(ptr, TRUE);
+	    parse_colors(ptr, NANO_REG_EXTENDED | REG_ICASE);
 	else if (strcasecmp(keyword, "linter") == 0)
-	    parse_linter(ptr);
+	    pick_up_name("linter", ptr, &live_syntax->linter);
 	else if (strcasecmp(keyword, "formatter") == 0)
 #ifndef DISABLE_SPELLER
-	    parse_formatter(ptr);
+	    pick_up_name("formatter", ptr, &live_syntax->formatter);
 #else
 	    ;
 #endif
+	else if (syntax_only)
+	    rcfile_error(N_("Command \"%s\" not allowed in included file"),
+					keyword);
+	else if (strcasecmp(keyword, "include") == 0)
+	    parse_includes(ptr);
+	else
 #endif /* !DISABLE_COLOR */
+	if (strcasecmp(keyword, "set") == 0)
+	    set = 1;
+	else if (strcasecmp(keyword, "unset") == 0)
+	    set = -1;
 	else if (strcasecmp(keyword, "bind") == 0)
 	    parse_binding(ptr, TRUE);
 	else if (strcasecmp(keyword, "unbind") == 0)
@@ -1180,12 +1039,9 @@ void parse_rcfile(FILE *rcstream
 	    rcfile_error(N_("Command \"%s\" not understood"), keyword);
 
 #ifndef DISABLE_COLOR
-	/* If we temporarily reset endsyntax to allow extending,
-	 * restore the value here. */
-	if (end_syn_save != NULL) {
-	    endsyntax = end_syn_save;
-	    end_syn_save = NULL;
-	}
+	/* If a syntax was extended, it stops at the end of the command. */
+	if (live_syntax != syntaxes)
+	    opensyntax = FALSE;
 #endif
 
 	if (set == 0)
@@ -1199,165 +1055,154 @@ void parse_rcfile(FILE *rcstream
 	option = ptr;
 	ptr = parse_next_word(ptr);
 
+	/* Find the just read name among the existing options. */
 	for (i = 0; rcopts[i].name != NULL; i++) {
-	    if (strcasecmp(option, rcopts[i].name) == 0) {
+	    if (strcasecmp(option, rcopts[i].name) == 0)
+		break;
+	}
+
+	if (rcopts[i].name == NULL) {
+	    rcfile_error(N_("Unknown option \"%s\""), option);
+	    continue;
+	}
+
 #ifdef DEBUG
-		fprintf(stderr, "parse_rcfile(): name = \"%s\"\n", rcopts[i].name);
+	fprintf(stderr, "parse_rcfile(): option name = \"%s\"\n", rcopts[i].name);
 #endif
-		if (set == 1) {
-		    if (rcopts[i].flag != 0)
-			/* This option has a flag, so it doesn't take an
-			 * argument. */
-			SET(rcopts[i].flag);
-		    else {
-			/* This option doesn't have a flag, so it takes
-			 * an argument. */
-			if (*ptr == '\0') {
-			    rcfile_error(
-				N_("Option \"%s\" requires an argument"),
+	/* First handle unsetting. */
+	if (set == -1) {
+	    if (rcopts[i].flag != 0)
+		UNSET(rcopts[i].flag);
+	    else
+		rcfile_error(N_("Cannot unset option \"%s\""), rcopts[i].name);
+	    continue;
+	}
+
+	/* If the option has a flag, it doesn't take an argument. */
+	if (rcopts[i].flag != 0) {
+	    SET(rcopts[i].flag);
+	    continue;
+	}
+
+	/* The option doesn't have a flag, so it takes an argument. */
+	if (*ptr == '\0') {
+	    rcfile_error(N_("Option \"%s\" requires an argument"),
 				rcopts[i].name);
-			    break;
-			}
-			option = ptr;
-			if (*option == '"')
-			    option++;
-			ptr = parse_argument(ptr);
+	    continue;
+	}
 
-			option = mallocstrcpy(NULL, option);
+	option = ptr;
+	if (*option == '"')
+	    option++;
+	ptr = parse_argument(ptr);
+
+	option = mallocstrcpy(NULL, option);
 #ifdef DEBUG
-			fprintf(stderr, "option = \"%s\"\n", option);
+	fprintf(stderr, "option argument = \"%s\"\n", option);
 #endif
-
-			/* Make sure option is a valid multibyte
-			 * string. */
-			if (!is_valid_mbstring(option)) {
-			    rcfile_error(
-				N_("Option is not a valid multibyte string"));
-			    break;
-			}
+	/* Make sure the option argument is a valid multibyte string. */
+	if (!is_valid_mbstring(option)) {
+	    rcfile_error(N_("Option is not a valid multibyte string"));
+	    continue;
+	}
 
 #ifndef DISABLE_COLOR
-			if (strcasecmp(rcopts[i].name, "titlecolor") == 0)
-			    specified_color_combo[TITLE_BAR] = option;
-			else if (strcasecmp(rcopts[i].name, "statuscolor") == 0)
-			    specified_color_combo[STATUS_BAR] = option;
-			else if (strcasecmp(rcopts[i].name, "keycolor") == 0)
-			    specified_color_combo[KEY_COMBO] = option;
-			else if (strcasecmp(rcopts[i].name, "functioncolor") == 0)
-			    specified_color_combo[FUNCTION_TAG] = option;
-			else
+	if (strcasecmp(rcopts[i].name, "titlecolor") == 0)
+	    specified_color_combo[TITLE_BAR] = option;
+	else if (strcasecmp(rcopts[i].name, "statuscolor") == 0)
+	    specified_color_combo[STATUS_BAR] = option;
+	else if (strcasecmp(rcopts[i].name, "keycolor") == 0)
+	    specified_color_combo[KEY_COMBO] = option;
+	else if (strcasecmp(rcopts[i].name, "functioncolor") == 0)
+	    specified_color_combo[FUNCTION_TAG] = option;
+	else
 #endif
 #ifndef DISABLE_OPERATINGDIR
-			if (strcasecmp(rcopts[i].name, "operatingdir") == 0)
-			    operating_dir = option;
-			else
+	if (strcasecmp(rcopts[i].name, "operatingdir") == 0)
+	    operating_dir = option;
+	else
 #endif
 #ifndef DISABLE_WRAPJUSTIFY
-			if (strcasecmp(rcopts[i].name, "fill") == 0) {
-			    if (!parse_num(option, &wrap_at)) {
-				rcfile_error(
-					N_("Requested fill size \"%s\" is invalid"),
-					option);
-				wrap_at = -CHARS_FROM_EOL;
-			    } else
-				free(option);
-			} else
+	if (strcasecmp(rcopts[i].name, "fill") == 0) {
+	    if (!parse_num(option, &wrap_at)) {
+		rcfile_error(N_("Requested fill size \"%s\" is invalid"),
+				option);
+		wrap_at = -CHARS_FROM_EOL;
+	    } else
+		free(option);
+	} else
 #endif
 #ifndef NANO_TINY
-			if (strcasecmp(rcopts[i].name,
-				"matchbrackets") == 0) {
-			    matchbrackets = option;
-			    if (has_blank_mbchars(matchbrackets)) {
-				rcfile_error(
-					N_("Non-blank characters required"));
-				free(matchbrackets);
-				matchbrackets = NULL;
-			    }
-			} else if (strcasecmp(rcopts[i].name,
-				"whitespace") == 0) {
-			    whitespace = option;
-			    if (mbstrlen(whitespace) != 2 ||
-				strlenpt(whitespace) != 2) {
-				rcfile_error(
-					N_("Two single-column characters required"));
-				free(whitespace);
-				whitespace = NULL;
-			    } else {
-				whitespace_len[0] =
-					parse_mbchar(whitespace, NULL,
-					NULL);
-				whitespace_len[1] =
-					parse_mbchar(whitespace +
+	if (strcasecmp(rcopts[i].name, "matchbrackets") == 0) {
+	    matchbrackets = option;
+	    if (has_blank_mbchars(matchbrackets)) {
+		rcfile_error(N_("Non-blank characters required"));
+		free(matchbrackets);
+		matchbrackets = NULL;
+	    }
+	} else if (strcasecmp(rcopts[i].name, "whitespace") == 0) {
+	    whitespace = option;
+	    if (mbstrlen(whitespace) != 2 || strlenpt(whitespace) != 2) {
+		rcfile_error(N_("Two single-column characters required"));
+		free(whitespace);
+		whitespace = NULL;
+	    } else {
+		whitespace_len[0] = parse_mbchar(whitespace, NULL, NULL);
+		whitespace_len[1] = parse_mbchar(whitespace +
 					whitespace_len[0], NULL, NULL);
-			    }
-			} else
+	    }
+	} else
 #endif
 #ifndef DISABLE_JUSTIFY
-			if (strcasecmp(rcopts[i].name, "punct") == 0) {
-			    punct = option;
-			    if (has_blank_mbchars(punct)) {
-				rcfile_error(
-					N_("Non-blank characters required"));
-				free(punct);
-				punct = NULL;
-			    }
-			} else if (strcasecmp(rcopts[i].name,
-				"brackets") == 0) {
-			    brackets = option;
-			    if (has_blank_mbchars(brackets)) {
-				rcfile_error(
-					N_("Non-blank characters required"));
-				free(brackets);
-				brackets = NULL;
-			    }
-			} else if (strcasecmp(rcopts[i].name,
-				"quotestr") == 0)
-			    quotestr = option;
-			else
+	if (strcasecmp(rcopts[i].name, "punct") == 0) {
+	    punct = option;
+	    if (has_blank_mbchars(punct)) {
+		rcfile_error(N_("Non-blank characters required"));
+		free(punct);
+		punct = NULL;
+	    }
+	} else if (strcasecmp(rcopts[i].name, "brackets") == 0) {
+	    brackets = option;
+	    if (has_blank_mbchars(brackets)) {
+		rcfile_error(N_("Non-blank characters required"));
+		free(brackets);
+		brackets = NULL;
+	    }
+	} else if (strcasecmp(rcopts[i].name, "quotestr") == 0)
+	    quotestr = option;
+	else
 #endif
 #ifndef NANO_TINY
-			if (strcasecmp(rcopts[i].name,
-				"backupdir") == 0)
-			    backup_dir = option;
-			else
+	if (strcasecmp(rcopts[i].name, "backupdir") == 0)
+	    backup_dir = option;
+	else
 #endif
 #ifndef DISABLE_SPELLER
-			if (strcasecmp(rcopts[i].name, "speller") == 0)
-			    alt_speller = option;
-			else
+	if (strcasecmp(rcopts[i].name, "speller") == 0)
+	    alt_speller = option;
+	else
 #endif
-			if (strcasecmp(rcopts[i].name,
-				"tabsize") == 0) {
-			    if (!parse_num(option, &tabsize) ||
-				tabsize <= 0) {
-				rcfile_error(
-					N_("Requested tab size \"%s\" is invalid"),
-					option);
-				tabsize = -1;
-			    } else
-				free(option);
-			} else
-			    assert(FALSE);
-		    }
+	if (strcasecmp(rcopts[i].name, "tabsize") == 0) {
+	    if (!parse_num(option, &tabsize) || tabsize <= 0) {
+		rcfile_error(N_("Requested tab size \"%s\" is invalid"),
+				option);
+		tabsize = -1;
+	    } else
+		free(option);
+	} else
+	    assert(FALSE);
+
 #ifdef DEBUG
-		    fprintf(stderr, "flag = %ld\n", rcopts[i].flag);
+	fprintf(stderr, "flag = %ld\n", rcopts[i].flag);
 #endif
-		} else if (rcopts[i].flag != 0)
-		    UNSET(rcopts[i].flag);
-		else
-		    rcfile_error(N_("Cannot unset option \"%s\""),
-			rcopts[i].name);
-		break;
-	    }
-	}
-	if (rcopts[i].name == NULL)
-	    rcfile_error(N_("Unknown option \"%s\""), option);
     }
 
 #ifndef DISABLE_COLOR
-    if (endsyntax != NULL && endcolor == NULL)
+    if (opensyntax && lastcolor == NULL)
 	rcfile_error(N_("Syntax \"%s\" has no color commands"),
-		endsyntax->desc);
+			live_syntax->name);
+
+    opensyntax = FALSE;
 #endif
 
     free(buf);
@@ -1421,7 +1266,7 @@ void do_rcfile(void)
 	/* Don't open directories, character files, or block files. */
 	if (stat(nanorc, &rcinfo) != -1) {
 	    if (S_ISDIR(rcinfo.st_mode) || S_ISCHR(rcinfo.st_mode) ||
-		S_ISBLK(rcinfo.st_mode))
+			S_ISBLK(rcinfo.st_mode))
 		rcfile_error(S_ISDIR(rcinfo.st_mode) ?
 			_("\"%s\" is a directory") :
 			_("\"%s\" is a device file"), nanorc);
@@ -1433,7 +1278,7 @@ void do_rcfile(void)
 	    /* Don't complain about the file's not existing. */
 	    if (errno != ENOENT)
 		rcfile_error(N_("Error reading %s: %s"), nanorc,
-			strerror(errno));
+				strerror(errno));
 	} else
 	    parse_rcfile(rcstream
 #ifndef DISABLE_COLOR
@@ -1447,8 +1292,7 @@ void do_rcfile(void)
 
     if (errors && !ISSET(QUIET)) {
 	errors = FALSE;
-	fprintf(stderr,
-		_("\nPress Enter to continue starting nano.\n"));
+	fprintf(stderr, _("\nPress Enter to continue starting nano.\n"));
 	while (getchar() != '\n')
 	    ;
     }
